class Alien extends Structure {
	constructor(game, type, doNotFadeIn){
		switch(type){
			case 0:
				super(game, game.textures.alien1, Structures.alien1);
				this.score = 100;
				break;
			
			case 1:
				super(game, game.textures.alien2, Structures.alien2);
				this.score = 150;
				this.r = Math.random() * 360;
				this.vr = 1.2;
				break;
			
			case 2:
				super(game, game.textures.alien3, Structures.alien3);
				this.score = 250;
				break;
			
			case 3:
				super(game, game.textures.alien4, Structures.alien4);
				this.score = 400;
				
				this.waitmin = 1.5;
				this.waitmax = 2.2;
				this.lenWait = Math.random() * (this.waitmax - this.waitmin) + this.waitmin;
				
				break;
			
			case 4:
				super(game, game.textures.alien5, Structures.alien5);
				this.score = 300;
				break;
			
			case 5:
				super(game, game.textures.alien6, Structures.alien6);
				game.playTrack("explode");
				this.score = 600;
				this.isBoss = true;
				this.invincible = true;
				
				// Why am I writing the boss fight code in the constructor?
				var that = this,
					alternate = false;
				
				this.beacons = [];
				this.drones = [ 1 ];
				
				for(var i=0; i<5; i++){
					var alien = new Alien(game, 6);
					game.scene.addObject(alien);
					this.beacons.push(alien);
					
					var angle = i/5 * 360,
						step = 140;
					
					alien.x = 330/2 + step *  Math.sin(angle/180*Math.PI),
					alien.y = 330/2 + step * -Math.cos(angle/180*Math.PI);
					alien.r = angle;
				}
				
				this.animations = [];
				this.attackEjectCount = 8;
				this.attackEject();
				
				var player = this.game.scene.player;
				
				var WW = this.game.animations.add(function(i){
					if(i !== 1) return;
					if(that.beacons.length === 0) return;
					if(that.beacons.length === 1) WW.time = WW.length = 8;
					
					if(that.beacons[0].invincible) return;
					
					that.selectedBeacon++;
					if(that.selectedBeacon >= that.beacons.length) that.selectedBeacon = 0;
					
					var beacon = that.beacons[that.selectedBeacon];
					var particle = new Reticle(that.game, player.x + (Math.random() * 2 - 1) * 30, player.y + (Math.random() * 2 - 1) * 30, 0, 0, 0, 0);
					that.game.scene.addParticle(particle);
					
					that.animations.push(that.game.animations.add(function(k){
						beacon.step = 140 - ease(k) * 80;
					}, null, 1));
					
					that.animations.push(that.game.animations.add(function(j){
						var desiredR = Math.atan2(particle.y - player.y, particle.x - player.x) * (180 / Math.PI) - 90,
							distance = Math.hypot(particle.x - player.x, particle.y - player.y),
							speed = 1/60 * 32;
						
						if(distance < speed){
							speed = distance;
						}
						
						particle.x += speed *  Math.sin(desiredR/180*Math.PI);
						particle.y += speed * -Math.cos(desiredR/180*Math.PI);
						
						beacon.overrideX = particle.x;
						beacon.overrideY = particle.y;
						beacon.overrideType = 0;
						
						if(j===1){
							beacon.overrideType = 1;
							that.game.playTrack("shoot");
							
							that.animations.push(that.game.animations.add(function(k){
								if(k===1){
									beacon.overrideType = 0;
									
									that.animations.push(that.game.animations.add(function(l){
										if(i===1){
											beacon.overrideType = null;
											beacon.overrideX = null;
											beacon.overrideY = null;
											
											beacon.step = 140 - ease(1 - l) * 80;
										}
									}, null, 1));
								}
							}, null, 2));
						}
					}, null, 3.5));
				}, null, 5, "infinite");
				
				this.selectedBeacon = 0;
				this.lenWait = 1.5;
				this.animations.push(WW);
			break;
			
			case 6:
				super(game, game.textures.alien7, Structures.alien7);
				this.step = 140;
				this.overrideX = null;
				this.overrideY = null;
				this.overrideType = null;
				this.invincible = true;
				this.score = 400;
				break;
		}
		
		this.type = type;
		
		// Fade in
		var that = this;
		this.opacity = 0;
		if(!doNotFadeIn) game.animations.add(function(i){ that.opacity = i; }, null, 4);
		
		// Weapons
		this.weapons = {
			"sniper": {
				projectile: PressureBolt,
				rounds: 1,
				cooldownTime: 0,
				reloadTime: 2
			},
			
			"laser": {
				projectile: PressureZip,
				rounds: 6,
				cooldownTime: 0.1,
				reloadTime: 0
			},
			
			"laser2": {
				projectile: PressureZip,
				rounds: 18,
				cooldownTime: 0.1,
				reloadTime: 0
			},
			
			"burst": {
				projectile: PressureBolt,
				rounds: 4,
				cooldownTime: 0.3,
				reloadTime: 2
			},
			
			"spray": {
				projectile: PressureBolt,
				rounds: 4,
				cooldownTime: 0.1,
				reloadTime: 2
			},
			
			"spray2": {
				projectile: PressureBolt,
				rounds: 2,
				cooldownTime: 0.2,
				reloadTime: 1.5
			},
			
			"bomb": {
				projectile: PressureMine,
				rounds: 3,
				cooldownTime: 0.2,
				reloadTime: 1
			},
			
			"charge": {
				projectile: PressureCharge,
				rounds: 1,
				cooldownTime: 0,
				reloadTime: this.lenWait
			}
		};
		
		this.weapon = {};
		this.animations = this.animations || [];
		
		this.useWeapon("burst");
		if(this.type===0) this.useWeapon("burst");
		if(this.type===1) this.useWeapon("spray");
		if(this.type===2) this.useWeapon("spray2");
		if(this.type===3) this.useWeapon("charge");
		if(this.type===4) this.useWeapon("bomb");
		if(this.type===6) this.useWeapon("laser");
		
		this.weapon.cooldownTime = Math.random() * this.weapons[this.weapon.type].reloadTime;
		this.isAIEnabled = true;
		
		if(this.type===6) this.weapon.cooldownTime = 0;
		// if(this.type===3) this.weapon.cooldownTime = 0;
	}
	
	
	useWeapon(type){
		this.weapon.type = type;
		this.weapon.rounds = this.weapons[type].rounds;
	}
	
	
	attackEject(){
		var that = this,
			lastDone = 0;
		
		this.attackEjectCount += 2;
		
		this.animations.push(this.game.animations.add(function(i){
			var nowDone = Math.floor(i * that.attackEjectCount);
			
			if(nowDone > lastDone){
				lastDone = nowDone;
				
				var alien = new Alien(that.game, 0, true);
				that.game.scene.addObject(alien);
				that.drones.push(alien);
				
				var wideAngle = 360,
				angle = that.r,
				step = 0,
				speed = 2;
				
				alien.x = that.x + step *  Math.sin(angle/180*Math.PI),
				alien.y = that.y + step * -Math.cos(angle/180*Math.PI);
				alien.vx = speed *  Math.sin(angle/180*Math.PI);
				alien.vy = speed * -Math.cos(angle/180*Math.PI);
				
				alien.r = that.r;
				alien.opacity = 1;
			}
		}, null, 0.25 * this.attackEjectCount));
	}
	
	
	update(delta, input){
		var actionForward = false, // input.keys[87],
			actionBackward = false, // input.keys[83],
			actionTiltLeft = false, // input.keys[65],
			actionTiltRight = false, // input.keys[68],
			actionLeft = false,
			actionRight = false,
			actionShoot = false,
			vDamp = 1.05,
			vrDamp = 1.05,
			rSpeed = 0.1,
			maxVel = 1,
			speed = 0;
		
		// AI
		if(this.isAIEnabled){
			var player = this.game.scene.player,
				distance = Math.hypot(this.x - player.x, this.y - player.y),
				desiredR = Math.atan2(this.y - player.y, this.x - player.x) * (180 / Math.PI) - 90;
			
			switch(this.type){
				// Small Ship
				case 0:
					var seekRadius = 70,
						avoidRadius = 30,
						seekAngle = 0,
						shootAngle = 10,
						peerRadius = 15;
					
					speed = 2.7;
					
					if(distance > seekRadius) actionForward = true;
					if(distance < avoidRadius) actionBackward = true;
					if(rotationInRange(desiredR, this.r, 0) === -1) actionTiltLeft = true;
					if(rotationInRange(desiredR, this.r, 0) === 1) actionTiltRight = true;
					if(rotationInRange(desiredR, this.r, shootAngle) === 0) actionShoot = true;
					
					peerRepel(this, peerRadius);
					
					break;
				
				// Medium Ship
				case 1:
					var targetX = 330/2,
						targetY = 330/2,
						targetR = Math.atan2(this.y - targetY, this.x - targetX) * (180 / Math.PI) - 90,
						targetDistance = Math.hypot(this.x - targetX, this.y - targetY),
						targetRadius = 165 - 15;
					
					speed = 1;
					vrDamp = 1;
					
					this.vx += delta * 0.6 *  Math.sin((targetR - 90)/180*Math.PI);
					this.vy += delta * 0.6 * -Math.cos((targetR - 90)/180*Math.PI);
					
					if(targetDistance > targetRadius){
						this.vx += delta * 0.6 *  Math.sin(targetR/180*Math.PI);
						this.vy += delta * 0.6 * -Math.cos(targetR/180*Math.PI);
					}
					
					else{
						if(rotationInRange(desiredR, this.r + 90, 10) === 0) actionShoot = true;
						if(rotationInRange(desiredR, this.r - 90, 10) === 0) actionShoot = true;
					}
					
					// peerRepel(this, 40);
					
					break;
				
				// Big Ship
				case 2:
					var seekRadius = 90,
						avoidRadius = 30,
						seekAngle = 0,
						shootAngle = 10,
						peerRadius = 20;
					
					speed = 0.5;
					rSpeed = 0.01;
					
					if(distance > seekRadius) actionForward = true;
					if(distance < avoidRadius) actionBackward = true;
					if(rotationInRange(desiredR, this.r, 0) === -1) actionTiltLeft = true;
					if(rotationInRange(desiredR, this.r, 0) === 1) actionTiltRight = true;
					if(rotationInRange(desiredR, this.r, shootAngle) === 0) actionShoot = true;
					
					peerRepel(this, peerRadius);
					
					break;
				
				// Ram Ship
				case 3:
					var lenMove = 0.5,
						ramspd = 110;
					
					var time = this.game.loop.frame % (60 * lenMove + 60 * this.lenWait) * delta * 60;
					
					vDamp = 1.1;
					actionShoot = true;
					
					if(rotationInRange(desiredR, this.r, 0) === -1) actionTiltLeft = true;
					if(rotationInRange(desiredR, this.r, 0) === 1) actionTiltRight = true;
					
					this.vx = (time/60 < lenMove ? 1 : 0.1) * delta * ramspd *  Math.sin(this.r/180*Math.PI);
					this.vy = (time/60 < lenMove ? 1 : 0.1) * delta * ramspd * -Math.cos(this.r/180*Math.PI);
					
					peerRepel(this, 15);
					
					break;
				
				// Burger Ship
				case 4:
					var targetX = 330/2,
						targetY = 330/2,
						targetR = Math.atan2(this.y - targetY, this.x - targetX) * (180 / Math.PI) - 90,
						targetDistance = Math.hypot(this.x - targetX, this.y - targetY),
						targetRadius = 165 - 15;
					
					speed = 1;
					
					if(targetDistance > targetRadius){
						this.vx += delta * 0.6 *  Math.sin((targetR + 90)/180*Math.PI);
						this.vy += delta * 0.6 * -Math.cos((targetR + 90)/180*Math.PI);
						
						this.vx += delta * 0.8 *  Math.sin(targetR/180*Math.PI);
						this.vy += delta * 0.8 * -Math.cos(targetR/180*Math.PI);
						
						this.r = targetR;
					}
					
					else{
						this.vr += (Math.random() * 2 - 1) * 0.2 * delta * 60;
						this.vx += delta * 2 *  Math.sin(this.r/180*Math.PI);
						this.vy += delta * 2 * -Math.cos(this.r/180*Math.PI);
						actionShoot = true;
					}
					
					break;
					
					// Boss ship
					case 5:
						var seekAngle = 30;
						
						rSpeed = 0.02;
						
						if(rotationInRange(desiredR, this.r - seekAngle, 0) === -1) actionTiltLeft = true;
						if(rotationInRange(desiredR, this.r + seekAngle, 0) === 1) actionTiltRight = true;
						
						for(var i=0; i<this.drones.length; i++) if((this.drones[i] === 1 && this.drones.length > 1) || this.drones[i].life <= 0) this.drones.splice(i, 1);
						for(var i=0; i<this.beacons.length; i++) if(this.beacons[i].life <= 0) this.beacons.splice(i, 1);
						
						
						if(!this.invincible){
							var lenMove = 0.5,
								ramspd = 110;
							
							var time = this.game.loop.frame % (60 * lenMove + 60 * this.lenWait) * delta * 60;
							
							vDamp = 1.1;
							rSpeed = 0.1;
							actionShoot = true;
							
							if(rotationInRange(desiredR, this.r, 0) === -1) actionTiltLeft = true;
							if(rotationInRange(desiredR, this.r, 0) === 1) actionTiltRight = true;
							
							this.vx = (time/60 < lenMove ? 1 : 0.1) * delta * ramspd *  Math.sin(this.r/180*Math.PI);
							this.vy = (time/60 < lenMove ? 1 : 0.1) * delta * ramspd * -Math.cos(this.r/180*Math.PI);
							
							peerRepel(this, 15);
							
							break
						}
						
						
						// Repel Player
						var player = this.game.scene.player,
							distance = Math.hypot(this.x - player.x, this.y - player.y),
							radius = 30;
						
						if(distance < radius){
							player.vx += (radius - distance) *  Math.sin(desiredR/180*Math.PI) / 100;
							player.vy += (radius - distance) * -Math.cos(desiredR/180*Math.PI) / 100;
						}
						
						
						// Drones
						if(this.drones.length === 0){
							this.drones.push(1);
							this.startBeaconLength = this.beacons.length;
							
							for(var i=0; i<this.beacons.length; i++){
								this.beacons[i].invincible = false;
								this.beacons[i].useWeapon("laser2");
							}
						}
						
						
						// Beacons
						if(this.beacons.length === 0){
							this.invincible = false;
						}
						
						else if(this.startBeaconLength - 2 >= this.beacons.length){
							this.startBeaconLength = 0;
							for(var i=0; i<this.beacons.length; i++) this.beacons[i].invincible = true;
							this.attackEject();
						}
					break;
					
					// Generator Ship
					case 6:
						this.r += delta * 5;
						this.x = 330/2 + this.step *  Math.sin(this.r/180*Math.PI);
						this.y = 330/2 + this.step * -Math.cos(this.r/180*Math.PI);
						
						if(!this.invincible) actionShoot = true;
						
						else {
							// Repel Player
							var player = this.game.scene.player,
								distance = Math.hypot(this.x - player.x, this.y - player.y),
								radius = 10;
							
							if(distance < radius){
								player.vx += (radius - distance) *  Math.sin(desiredR/180*Math.PI) / 100;
								player.vy += (radius - distance) * -Math.cos(desiredR/180*Math.PI) / 100;
							}
						}
						
					break;
			}
		}
		
		
		function rotationInRange(a, b, range){
			var d = mod(a - b + 180, 360) - 180;
			return (d >= -range && d <= range) ? 0 : (
				d < -range ? -1 : 1
			);
		}
		
		
		function peerRepel(that, peerRadius){
			var index = Infinity;
			
			for(var i=0; i<that.game.scene.objects.length; i++){
				var object = that.game.scene.objects[i];
				if(object.constructor !== Alien) continue;
				
				if(object===that){
					index = i;
					continue;
				}
				
				var distance = Math.hypot(that.x - object.x, that.y - object.y);
				
				if(distance < peerRadius){
					if(index > i){
						actionLeft = true;
						actionBackward = true;
					}
					
					else{
						actionRight = true;
						actionForward = true;
					}
					
					return;
				}
			}
		}
		
		
		// Input Stuff
		if(actionForward && actionBackward) actionForward = actionBackward = false;
		if(actionTiltLeft && actionTiltRight) actionTiltLeft = actionTiltRight = false;
		
		
		// Apply Controls
		if(actionForward){
			this.vx += delta * speed *  Math.sin(this.r/180*Math.PI);
			this.vy += delta * speed * -Math.cos(this.r/180*Math.PI);
		}
		
		if(actionBackward){
			this.vx += delta * -speed *  Math.sin(this.r/180*Math.PI);
			this.vy += delta * -speed * -Math.cos(this.r/180*Math.PI);
		}
		
		if(actionLeft){
			this.vx += delta * speed / 2 *  Math.sin((this.r - 90)/180*Math.PI);
			this.vy += delta * speed / 2 * -Math.cos((this.r - 90)/180*Math.PI);
		}
		
		if(actionRight){
			this.vx += delta * speed / 2 *  Math.sin((this.r + 90)/180*Math.PI);
			this.vy += delta * speed / 2 * -Math.cos((this.r + 90)/180*Math.PI);
		}
		
		if(actionTiltLeft){
			this.vr -= delta * 60 * rSpeed;
		}
		
		if(actionTiltRight){
			this.vr += delta * 60 * rSpeed;
		}
		
		if(this.weapon.rounds===0){
			this.weapon.rounds = this.weapons[this.weapon.type].rounds;
			this.weapon.cooldownTime = this.weapons[this.weapon.type].reloadTime;
		}
		
		if(actionShoot && this.weapon.cooldownTime <= 0){
			var direction = this.spellCount % 2 === 0 ? 1 : -1;
			
			if(this.type !== 6) this.game.playTrack("shoot");
			else if(this.weapon.cooldownTime === this.weapon.cooldownTimeMax) this.game.playTrack("shoot");
			
			switch(this.type){
				case 6:
					var offsetR = 6,
						angle = 14;
					
					if(this.overrideType === null){
						var projectile = new this.weapons["laser"].projectile(this.game, this.x, this.y, this.r + 180 + angle + offsetR, direction, this.weapon.type==="laser2");
						this.game.scene.addObject(projectile);
						projectile.parent = this;
					}
					
					else{
						var desiredR = Math.atan2(this.y - this.overrideY, this.x - this.overrideX) * (180 / Math.PI) - 90,
							projectile = new this.weapons["laser"].projectile(this.game, this.x, this.y, desiredR, direction, this.overrideType === 0);
						
						this.game.scene.addObject(projectile);
						projectile.parent = this;
					}
					
					break;
				
				case 5:
					var projectile = new this.weapons[this.weapon.type].projectile(this.game, this.x, this.y, this.r - 10, direction);
					this.game.scene.addObject(projectile);
					projectile.parent = this;
					
					var projectile = new this.weapons[this.weapon.type].projectile(this.game, this.x, this.y, this.r - 5, direction);
					this.game.scene.addObject(projectile);
					projectile.parent = this;
					
					var projectile = new this.weapons[this.weapon.type].projectile(this.game, this.x, this.y, this.r, direction);
					this.game.scene.addObject(projectile);
					projectile.parent = this;
					
					var projectile = new this.weapons[this.weapon.type].projectile(this.game, this.x, this.y, this.r + 5, direction);
					this.game.scene.addObject(projectile);
					projectile.parent = this;
					
					var projectile = new this.weapons[this.weapon.type].projectile(this.game, this.x, this.y, this.r + 10, direction);
					this.game.scene.addObject(projectile);
					projectile.parent = this;
					
					break;
				
				case 4:
					var offsetX = -8 *  Math.sin(this.r/180*Math.PI),
						offsetY = -8 * -Math.cos(this.r/180*Math.PI);
					
					var projectile = new this.weapons[this.weapon.type].projectile(this.game, this.x + offsetX, this.y + offsetY, this.r, direction);
					this.game.scene.addObject(projectile);
					projectile.parent = this;
					break;
				
				case 3:
					var projectile = new this.weapons[this.weapon.type].projectile(this.game, this.x, this.y, this.r - 10, direction);
					this.game.scene.addObject(projectile);
					projectile.parent = this;
					
					var projectile = new this.weapons[this.weapon.type].projectile(this.game, this.x, this.y, this.r, direction);
					this.game.scene.addObject(projectile);
					projectile.parent = this;
					
					var projectile = new this.weapons[this.weapon.type].projectile(this.game, this.x, this.y, this.r + 10, direction);
					this.game.scene.addObject(projectile);
					projectile.parent = this;
					break;
				
				case 1:
					var projectile = new this.weapons[this.weapon.type].projectile(this.game, this.x, this.y, this.r + 90, direction);
					this.game.scene.addObject(projectile);
					projectile.parent = this;
					
					projectile = new this.weapons[this.weapon.type].projectile(this.game, this.x, this.y, this.r - 90, direction);
					this.game.scene.addObject(projectile);
					projectile.parent = this;
					break;
				
				default:
					var projectile = new this.weapons[this.weapon.type].projectile(this.game, this.x, this.y, this.r, direction);
					this.game.scene.addObject(projectile);
					projectile.parent = this;
			}
			
			this.weapon.cooldownTime = this.weapons[this.weapon.type].cooldownTime;
			this.weapon.rounds--;
			
			if(this.type === 6 && this.weapon.rounds <= 0){
				if(this.weapon.type === "laser" ) this.useWeapon("laser2");
				else if(this.weapon.type === "laser2") this.useWeapon("laser");
			}
		}
		
		
		// Physics
		this.x += this.vx * delta * 60;
		this.y += this.vy * delta * 60;
		this.r += this.vr * delta * 60;
		
		this.vx /= vDamp;
		this.vy /= vDamp;
		this.vr /= vrDamp;
		
		this.r %= 360;
		
		if(this.life<=0){
			this.game.scene.addParticle(new ScoreParticle(this.game, this.x, this.y, "+" + this.score));
			this.game.scene.setRule("score", this.game.scene.getRule("score") + this.score);
			
			while(this.animations[0]) this.game.animations.remove(this.animations.splice(0, 1)[0]);
			
			var hs = +atob(localStorage.SpaceWizard || "MA=="), sc = this.game.scene.getRule("score");
			if(sc > hs) localStorage.SpaceWizard = btoa(sc);
			
			switch(this.type){
				case 6:
					var offsetR = Math.random() * 360;
					
					for(var i=0; i<15; i++){
						var projectile = new PressureBolt(this.game, this.x, this.y, this.r + offsetR + (i/15 * 360), null);
						this.game.scene.addObject(projectile);
						projectile.parent = this;
					}
					
					break;
					
				case 5:
					for(var i=0; i<15; i++){
						var offsetX = this.scale * 4 + (Math.random() * 2 - 1) * 15,
							offsetY = this.scale * 5 + (Math.random() * 2 - 1) * 30,
							
							x = this.x,
							y = this.y;
						
						x += offsetX *  Math.sin((this.r - 90)/180*Math.PI);
						y += offsetX * -Math.cos((this.r - 90)/180*Math.PI);
						
						x += offsetY *  Math.sin(this.r/180*Math.PI);
						y += offsetY * -Math.cos(this.r/180*Math.PI);
						
						var particle = new Explosion(this.game, this.game.textures.projectiles, this.game.atlas.particles.explosion.green, x, y, this.r, this.vx, this.vy, this.vr);
						
						particle.maxLife = Math.random() + 0.5;
						particle.life = particle.maxLife;
						this.game.scene.addParticle(particle)
					}
					
					this.game.shake(0.5, 1.5);
					
					break;
				
				case 4:
				case 3:
					this.game.scene.addParticle(new Explosion(this.game, this.game.textures.projectiles, this.game.atlas.particles.explosion.green, this.x, this.y, this.r, this.vx, this.vy, this.vr))
					this.game.shake(0.2, 0.5);
					break;
				
				default:
					this.game.scene.addParticle(new Explosion(this.game, this.game.textures.projectiles, this.game.atlas.particles.explosion.green, this.x, this.y, this.r, this.vx, this.vy, this.vr))
					this.game.shake(0.1, 0.2);
			}
			
			this.game.playTrack("explode");
		}
		
		this.weapon.cooldownTime -= delta;
		return this.life;
	}
}
